; asm6

; iNES header
    .db "NES",$1A                               ; iNES signature
    .db $04                                     ; Number of 0x4000 PRG ROM banks
    .db nChr                                    ; Number of 0x2000 CHR ROM banks
    .db $41                                     ; Mirroring and mapper lower nibble
    .db $00                                     ; mapper upper nibble

    .db $00,$00,$00,$00,$00,$00,$00,$00         ; unused 

; Zero page variables
    .base $0000
    .enum $0000
    include code\zeroPage.asm
    .ende

; RAM variables
    .base $0200
    .enum $0200
    include code\ram.asm
    .ende

; Include constants, macros
include code\macros.asm                         ; macros
include code\nesRegisters.asm                   ; various NES-specific constants
include code\constants.asm                      ; other constants
include code\constauto.asm                      ; automatically generated constants
include code\banks.asm                          ; Bank asignments

; Now we are in the fixed bank at $c000
include code\joypad.asm
include code\printer.asm                        ; Screen print routine
include code\text.asm
include code\palettes.asm
include code\loadPalette.asm
include code\nametable.asm

loadNameTable:
    ptr = $00 ; zero page pointer at $00, $01


    bit PPUSTATUS                  ; Reset address latch flip-flop
    lda #$20
    sta PPUADDR                 ; Write address high byte to PPU
    lda #$00
    sta PPUADDR                 ; Write address low byte to PPU
    
    
    lda #<nametable_data
    sta ptr+0
    lda #>nametable_data
    sta ptr+1
    ldx #4 ; do this loop 4 times
    ldy #0
loadNTLoop:
    lda (ptr), Y
    sta $2007
    iny
    bne loadNTLoop
    dex
    beq + ; finished if X = 0
    inc ptr+1 ; ptr = ptr + 256
    jmp loadNTLoop ; loop again: Y = 0, X -= 1, ptr += 256
+
    rts

NMI:
    pha                         ; push a,x,y on to the stack
    txa
    pha
    tya
    pha
    
    lda skipNMI                 ; This is for when we want to skip most of
    bne skipNMIStuff            ; the NMI, like when loading a lot of stuff.
    
    lda #$0c
    jsr setRightCHRDirect       ; Set CHR to hud tiles.  We use "Direct" here because it's temporary.

    ; hud irq stuff
    ;lda #$1e                    ; This value only matters for $c000 and $c001
    lda #$40
    sta $e000                   ; Acknowledge any pending interrupts
    sta $c000                   ; Set IRQ counter to number of scanlines to wait
    sta $c001                   ; Write it again to the IRQ counter latch
    sta $e001                   ; Enable IRQ
    
    lda #$00
    sta splitNum
    
    lda #$02                    ; transfer sprites from $0200 to the ppu
    sta OAMDMA
    
    jsr writeBufferToPPU
    jsr handleMessageQueue
    
    jsr loadPalette
    
    lda #$90
    sta PPUCTRL
    
    ; Have to reset Scroll since some of this stuff corrupts it.
    ; Also, since we have a hud we'll set it to the hud scroll values.
    ; The game scroll values will be set in the irq.
    ;
    bit PPUSTATUS               ; Reset address latch flip-flop
    lda #$00                    ; reset PPU address to avoid glitches.
    sta PPUADDR
    sta PPUADDR
    ;lda #$fa                    ; Set this to give a little bit of padding to the hud text.  Neat!
    lda #$00
    lda altScrollX_hi
    sta PPUSCROLL
    lda #$00
    sta PPUSCROLL
    
skipNMIStuff:
    
    lda #$01
    sta vblanked
    
    pla                         ; pull y,x,a off the stack
    tay
    pla
    tax
    pla
    
    rti
IRQ:
    pha                         ; push a,x,y on to the stack
    txa
    pha
    tya
    pha

    sta $e000                   ; Acknowledge the IRQ and disable.
    
    bit PPUSTATUS               ; Reset address latch flip-flop

    lda splitNum
    beq +                       ; branch to the + and put 0 in the scroll for x

    lda scrollX_hi              ; Set our scroll values here after the hud split.
+
    sta PPUSCROLL
    lda scrollY_hi
    sta PPUSCROLL
    
    lda splitNum
    cmp #$01                    ; split #2, no more irqs
    beq +
    
    lda #$20
    sta $e000                   ; Acknowledge any pending interrupts
    sta $c000                   ; Set IRQ counter to number of scanlines to wait
    sta $c001                   ; Write it again to the IRQ counter latch
    sta $e001                   ; Enable IRQ
    
    inc splitNum
+
    
    jsr setCHR                  ; Set CHR banks based on CHR0, CHR1, etc.
    
    pla                         ; pull y,x,a off the stack
    tay
    pla
    tax
    pla

    rti

; The reset vector and mmc3 initialize code must point into $E000-$FFFF
; because the state of the mapper at power-on is unspecified.
.pad $e000
include code\mmc3.asm

Reset:
    include code\reset.asm
main:
    lda #$01                    ; set mirroring to horizontal
    jsr setMirroring
    
    lda #$40                    ; disable frame counter to make irq work
    sta $4017
    
    cli                         ; Clear interrupt Disable Bit (so we can use irqs)
    
    lda #$01
    sta skipNMI                 ; skip (most of) NMI code.
    
    LDA #$00
    sta PPUMASK
    jsr waitframe
    
    lda #$00
    jsr setLeftCHR
    lda #$0c
    jsr setRightCHR

    ;jsr drawTitle
    
    lda #$00
    sta paletteSet
    
    jsr loadNameTable
    
;    lda #$03
;    jsr addMessageToQueue
;    jsr handleMessageQueue
;    lda #$04
;    jsr addMessageToQueue
;    jsr handleMessageQueue
;    lda #$05
;    jsr addMessageToQueue
;    jsr handleMessageQueue
;    lda #$06
;    jsr addMessageToQueue
;    jsr handleMessageQueue
    
;    lda #$01
;    jsr addMessageToQueue
;    jsr waitframe
    
    bit PPUSTATUS               ; Reset address latch flip-flop
    lda #$00                    ; reset PPU address to avoid glitches.
    sta PPUADDR
    sta PPUADDR
    lda scrollX_hi
    sta PPUSCROLL               ; reset Scroll since some of this stuff corrupts it.
    lda scrollY_hi
    sta PPUSCROLL
    
    jsr waitframe
    lda #$1e                    ; Turn on rendering
    sta PPUMASK
    
    lda #$00
    sta skipNMI
    
;    lda #SELECTED_PALETTE
;    sta palettes
    
    lda #$1e                    ; Turn on rendering
    sta PPUMASK
    
mainLoop:
    jsr waitframe
    jsr readJoy
    
    lda paused
    beq +
    lda buttonsPress
    cmp #JOY_SELECT             ; Press select while paused to reset
    bne +
    jmp Reset
+
    
    lda buttonsPress
    cmp #JOY_START              ; Pause
    bne +
    lda paused
    eor #$01
    sta paused
    bne +
+
    lda paused
    bne mainLoop
    
    
    lda scrollX                 ; Scroll the level
    clc
    adc #scrollSpeed
    sta scrollX
    lda scrollX_hi
    adc #scrollSpeed_hi
    sta scrollX_hi
    
    lda scrollX_page
    adc #$00
    sta scrollX_page
    
    lda altScrollX                 ; Scroll the level
    clc
    adc #altScrollSpeed
    sta altScrollX
    lda altScrollX_hi
    adc #altScrollSpeed_hi
    sta altScrollX_hi
    
    lda altScrollX_page
    adc #$00
    sta altScrollX_page
    
    lda buttonsPress
    cmp #JOY_SELECT
    bne +
    inc paletteSet
    lda paletteSet
    and #$03
    sta paletteSet
+
    
    jmp mainLoop

writeBufferToPPU:
    lda buffer1Offset
    beq bufferDone      ; If the buffer offset is 0, it's empty.  Skip to end.
    
    ldx #$00            ; Initialize x for our buffer index
bufferLoop1:
    ldy #$00
    
    bit PPUSTATUS       ; Reset address latch flip-flop
    lda buffer1,x       ; Load first byte from our buffer chunk
    sta PPUADDR         ; Write address high byte to PPU
    inx
    lda buffer1,x       ; Load first byte from our buffer chunk
    sta PPUADDR         ; Write address low byte to PPU
    inx
    
    lda buffer1,x       ; Get length of this chunk
    tay                 ; Store in y to indicate when this chunk is done
    inx
bufferLoop2:
    lda buffer1,x       ; Get next byte from buffer
    sta PPUDATA         ; Store byte in PPU
    inx                 ; Increase buffer index
    dey                 ; Decrease y
    bne bufferLoop2     ; If y is not 0, we haven't reached the end of this chunk
    
    cpx buffer1Offset   ; Check if our buffer index in x matches the stored buffer offset
    bne bufferLoop1     ; If not, read another chunk
    
    lda #$00            ; Reset our stored buffer offset
    sta buffer1Offset
bufferDone:
    rts

waitframe:
    lda #$00
    sta vblanked
waitloop:
    lda vblanked
    beq waitloop
    rts

    .org $fffa

    .dw NMI
    .dw Reset
    .dw IRQ

    include code/chrlist.asm
